# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

# Assume unicode literals as per Python 3
from __future__ import unicode_literals

import binascii
import random
from copy import copy, deepcopy

import six

import hszinc
from hszinc.datatypes import XStr, Uri, Bin, MARKER, NA, REMOVE
from hszinc.pintutil import to_haystack, to_pint
from .pint_enable import _enable_pint

from nose.tools import eq_


def check_singleton_deepcopy(S):
    orig_dict = {'some_value': S}
    copy_dict = deepcopy(orig_dict)
    assert copy_dict['some_value'] is S


def test_marker_deepcopy():
    check_singleton_deepcopy(hszinc.MARKER)


def test_marker_hash():
    assert hash(hszinc.MARKER) == hash(hszinc.MARKER.__class__)


def test_remove_deepcopy():
    check_singleton_deepcopy(hszinc.REMOVE)


def test_remove_hash():
    assert hash(hszinc.REMOVE) == hash(hszinc.REMOVE.__class__)


def check_singleton_copy(S):
    assert copy(S) is S


def test_marker_copy():
    check_singleton_copy(hszinc.MARKER)


def test_remove_copy():
    check_singleton_copy(hszinc.REMOVE)


def test_ref_notref_eq():
    r1 = hszinc.Ref(name='a.ref')
    r2 = 'not.a.ref'
    assert r1 is not r2
    assert not (r1 == r2)


def test_ref_notref_ne():
    r1 = hszinc.Ref(name='a.ref')
    r2 = 'not.a.ref'
    assert r1 is not r2
    assert r1 != r2


def test_ref_simple_eq():
    r1 = hszinc.Ref(name='a.ref')
    r2 = hszinc.Ref(name='a.ref')
    assert r1 is not r2
    assert r1 == r2


def test_ref_simple_neq_id():
    r1 = hszinc.Ref(name='a.ref')
    r2 = hszinc.Ref(name='another.ref')
    assert r1 is not r2
    assert r1 != r2


def test_ref_mixed_neq():
    r1 = hszinc.Ref(name='a.ref')
    r2 = hszinc.Ref(name='a.ref', value='display text')
    assert r1 is not r2
    assert r1 != r2


def test_ref_withdis_eq():
    r1 = hszinc.Ref(name='a.ref', value='display text')
    r2 = hszinc.Ref(name='a.ref', value='display text')
    assert r1 is not r2
    assert r1 == r2


def test_ref_withdis_neq_id():
    r1 = hszinc.Ref(name='a.ref', value='display text')
    r2 = hszinc.Ref(name='another.ref', value='display text')
    assert r1 is not r2
    assert r1 != r2


def test_ref_withdis_neq_dis():
    r1 = hszinc.Ref(name='a.ref', value='display text')
    r2 = hszinc.Ref(name='a.ref', value='different display text')
    assert r1 is not r2
    assert r1 != r2


def test_ref_hash():
    assert hash(hszinc.Ref(name='a.ref', value='display text')) == \
           hash('a.ref') ^ hash('display text') ^ hash(True)


def test_ref_std_method():
    if six.PY2:
        assert str(hszinc.Ref(name='a.ref', value='display text')) == '@a.ref u\'display text\''
    else:
        assert str(hszinc.Ref(name='a.ref', value='display text')) == '@a.ref \'display text\''


def test_qty_unary_ops():
    # How to run the test: check the result
    # applied to the Quantity object matches what
    # would be returned for the same operation applied
    # to the raw value.
    def _check_qty_op(pint_en, fn, *vals):
        _enable_pint(pint_en)
        for v in vals:
            q = hszinc.Quantity(v)
            assert fn(q) == fn(q.value)

    # Try this both without, and with, pint enabled
    for pint_en in (False, True):
        # These work for floats
        for fn in (lambda v: int(v),
                   lambda v: complex(v),
                   lambda v: float(v),
                   lambda v: -v,
                   lambda v: +v,
                   lambda v: abs(v)):
            yield _check_qty_op, pint_en, fn, 123.45, -123.45

        # These work for integers
        for fn in (lambda v: oct(v),
                   lambda v: hex(v),
                   lambda v: v.__index__(),
                   lambda v: int(v),
                   lambda v: complex(v),
                   lambda v: float(v),
                   lambda v: -v,
                   lambda v: +v,
                   lambda v: abs(v),
                   lambda v: ~v):
            yield _check_qty_op, pint_en, fn, 123, -123

        # This only works with Python 2.
        if six.PY2:  # pragma: no cover
            yield _check_qty_op, pint_en, lambda v: long(v), 123.45, -123.45


def test_qty_hash():
    def _check_qty_hash(pint_en):
        # Test that independent copies hash to the same value
        def _check_hash(val, unit=None):
            a = hszinc.Quantity(val, unit=unit)
            b = hszinc.Quantity(val, unit=unit)

            assert a is not b
            assert hash(a) == hash(b)

        _enable_pint(pint_en)

        _check_hash(123.45)
        _check_hash(-123.45)
        _check_hash(12345)
        _check_hash(-12345)
        _check_hash(50, 'Hz')

    for pint_en in (False, True):
        yield _check_qty_hash, pint_en


def test_qty_binary_ops():
    def _check_qty_op(pint_en, fn, a, b):
        _enable_pint(pint_en)
        qa = hszinc.Quantity(a)
        qb = hszinc.Quantity(b)

        # Reference value
        ref = fn(a, b)

        assert fn(qa, qb) == ref
        assert fn(qa, b) == ref
        assert fn(a, qb) == ref

    for pint_en in (False, True):
        # Try some float values
        floats = (1.12, 2.23, -4.56, 141.2, -399.5)
        for a in floats:
            for b in floats:
                if a == b:
                    continue

                for fn in (lambda a, b: a + b,
                           lambda a, b: a - b,
                           lambda a, b: a * b,
                           lambda a, b: a / b,
                           lambda a, b: a // b,
                           lambda a, b: a % b,
                           lambda a, b: divmod(a, b),
                           lambda a, b: a < b,
                           lambda a, b: a <= b,
                           lambda a, b: a == b,
                           lambda a, b: a != b,
                           lambda a, b: a >= b,
                           lambda a, b: a > b):
                    yield _check_qty_op, pint_en, fn, a, b

        # Exponentiation, we can't use all the values above
        # as some go out of dates_range.
        small_floats = tuple(filter(lambda f: abs(f) < 10, floats))
        for a in small_floats:
            for b in small_floats:
                if a == b:
                    continue

                # Python2 won't allow raising negative numbers
                # to a fractional power
                if a < 0:
                    continue

                yield _check_qty_op, pint_en, lambda a, b: a ** b, a, b

        # Try some integer values
        ints = (1, 2, -4, 141, -399, 0x10, 0xff, 0x55)
        for a in ints:
            for b in ints:
                if a == b:
                    continue

                for fn in (lambda a, b: a + b,
                           lambda a, b: a - b,
                           lambda a, b: a * b,
                           lambda a, b: a / b,
                           lambda a, b: a // b,
                           lambda a, b: a % b,
                           lambda a, b: divmod(a, b),
                           lambda a, b: a & b,
                           lambda a, b: a ^ b,
                           lambda a, b: a | b,
                           lambda a, b: a < b,
                           lambda a, b: a <= b,
                           lambda a, b: a == b,
                           lambda a, b: a != b,
                           lambda a, b: a >= b,
                           lambda a, b: a > b):
                    yield _check_qty_op, pint_en, fn, a, b

                if b >= 0:
                    for fn in (lambda a, b: a << b,
                               lambda a, b: a >> b):
                        yield _check_qty_op, pint_en, fn, a, b

        # Exponentiation, we can't use all the values above
        # as some go out of dates_range.
        small_ints = tuple(filter(lambda f: abs(f) < 10, ints))
        for a in small_ints:
            for b in small_ints:
                if a == b:
                    continue

                yield _check_qty_op, pint_en, lambda a, b: a ** b, a, b


def test_qty_cmp():
    def _check_qty_cmp(pint_en):
        if 'cmp' not in set(locals().keys()):
            def cmp(a, b):
                return a.__cmp__(b)

        _enable_pint(pint_en)

        a = hszinc.Quantity(-3)
        b = hszinc.Quantity(432)
        c = hszinc.Quantity(4, unit='A')
        d = hszinc.Quantity(10, unit='A')
        e = hszinc.Quantity(12, unit='V')

        assert cmp(a, b) < 0
        assert cmp(b, a) > 0
        assert cmp(a, hszinc.Quantity(-3)) == 0
        assert cmp(c, d) < 0
        assert cmp(d, c) > 0
        assert cmp(c, hszinc.Quantity(4, unit='A')) == 0

        try:
            cmp(c, e)
        except TypeError as ex:
            assert str(ex) == 'Quantity units differ: A vs V'

    for pint_en in (False, True):
        yield _check_qty_cmp, pint_en


def test_qty_std_method():
    def _check_qty_cmp(pint_en):
        r = repr(hszinc.Quantity(4, unit='A'))
        if six.PY2:
            assert (r == 'BasicQuantity(4, u\'A\')') or (r == 'PintQuantity(4, u\'A\')')
        else:
            assert (r == 'BasicQuantity(4, \'A\')') or (r == 'PintQuantity(4, \'A\')')
        assert str(hszinc.Quantity(4, unit='A')) == '4 A'

    yield _check_qty_cmp, True
    yield _check_qty_cmp, False


class MyCoordinate(object):
    """
    A dummy class that can compare itself to a Coordinate from hszinc.
    """

    def __init__(self, lat, lng):
        self.lat = lat
        self.lng = lng

    def __eq__(self, other):
        if isinstance(other, hszinc.Coordinate):
            return (self.lat == other.latitude) \
                   and (self.lng == other.longitude)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, hszinc.Coordinate):
            return (self.lat != other.latitude) \
                   and (self.lng != other.longitude)
        return NotImplemented


def test_coord_eq():
    assert hszinc.Coordinate(latitude=33.77, longitude=-77.45) \
           == hszinc.Coordinate(latitude=33.77, longitude=-77.45)


def test_coord_eq_notcoord():
    assert not (hszinc.Coordinate(latitude=33.77, longitude=-77.45) \
                == (33.77, -77.45))


def test_coord_eq_mycoord():
    hsc = hszinc.Coordinate(latitude=33.77, longitude=-77.45)
    mc = MyCoordinate(33.77, -77.45)
    assert hsc == mc
    assert mc == hsc


def test_coord_ne():
    assert hszinc.Coordinate(latitude=-33.77, longitude=77.45) \
           != hszinc.Coordinate(latitude=33.77, longitude=-77.45)


def test_coord_ne_notcoord():
    assert (hszinc.Coordinate(latitude=33.77, longitude=-77.45) \
            != (33.77, -77.45))


def test_coord_ne_mycoord():
    hsc = hszinc.Coordinate(latitude=33.77, longitude=-77.45)
    mc = MyCoordinate(-33.77, 77.45)
    assert hsc != mc
    assert mc != hsc


def test_coord_hash():
    assert hash(hszinc.Coordinate(latitude=33.77, longitude=-77.45)) == \
           hash(33.77) ^ hash(-77.45)


def test_coord_default_method():
    coord = hszinc.Coordinate(latitude=33.77, longitude=-77.45)
    ref_str = u'33.770000° lat -77.450000° long'
    if six.PY2:
        ref_str = ref_str.encode('utf-8')

    eq_(repr(coord), 'Coordinate(33.77, -77.45)')
    eq_(str(coord), ref_str)


def test_xstr_hex():
    assert XStr("hex", "deadbeef").data == b'\xde\xad\xbe\xef'
    barray = bytearray(random.getrandbits(8) for _ in range(10))
    assert barray == hszinc.XStr("hex", binascii.hexlify(barray).decode("ascii")).data


def test_xstr_other():
    assert (XStr("other", "hello word").data == "hello word")
    barray = bytearray(random.getrandbits(8) for _ in range(10))
    assert barray == hszinc.XStr("other", barray).data


def test_xstr_b64():
    assert XStr("b64", '3q2+7w==').data == b'\xde\xad\xbe\xef'
    barray = bytearray(random.getrandbits(8) for _ in range(10))
    assert barray == hszinc.XStr("b64", binascii.b2a_base64(barray)).data


def test_xstr_equal():
    assert XStr("hex", "deadbeef") == XStr("b64", '3q2+7w==')


def test_uri():
    uri = Uri("abc")
    assert uri == Uri("abc")
    if six.PY2:
        assert repr(uri) == 'Uri(u\'abc\')'
    else:
        assert repr(uri) == 'Uri(\'abc\')'
    assert str(uri) == 'abc'


def test_bin():
    bin = Bin("text/plain")
    if six.PY2:
        assert repr(bin) == 'Bin(u\'text/plain\')'
    else:
        assert repr(bin) == 'Bin(\'text/plain\')'
    assert str(bin) == 'text/plain'


def test_marker():
    assert repr(MARKER) == 'MARKER'


def test_na():
    assert repr(NA) == 'NA'


def test_remove():
    assert repr(REMOVE) == 'REMOVE'


def test_to_haystack():
    assert to_haystack('/h') == u''
    assert to_haystack(u'foot ** 3') == u'cubic_foot'
    assert to_haystack(u'deg') == u'\N{DEGREE SIGN}'

def test_to_pint():
    assert to_pint(u'\N{DEGREE SIGN}') == 'deg'
    assert to_pint('cubic_foot') == u'cubic foot'



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

import datetime
import json

import pytz

import hszinc
from .test_parser import SIMPLE_EXAMPLE, SIMPLE_EXAMPLE_JSON, \
    METADATA_EXAMPLE_JSON

# The metadata example is a little different, as we generate the grid without
# spaces around the commas.
METADATA_EXAMPLE = '''ver:"2.0" database:"test" dis:"Site Energy Summary"
siteName dis:"Sites",val dis:"Value" unit:"kW"
"Site 1",356.214kW
"Site 2",463.028kW
'''


def make_simple_grid(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    grid.column['firstName'] = {}
    grid.column['bday'] = {}
    grid.extend([
        {
            'firstName': 'Jack',
            'bday': datetime.date(1973, 7, 23),
        },
        {
            'firstName': 'Jill',
            'bday': datetime.date(1975, 11, 15),
        },
    ])
    return grid


def test_simple():
    grid = make_simple_grid()
    grid_str = hszinc.dump(grid)
    assert grid_str == SIMPLE_EXAMPLE


def test_simple_json():
    grid = make_simple_grid()
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == SIMPLE_EXAMPLE_JSON


def make_metadata_grid(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    grid.metadata['database'] = 'test'
    grid.metadata['dis'] = 'Site Energy Summary'
    grid.column['siteName'] = {'dis': 'Sites'}
    grid.column['val'] = hszinc.MetadataObject()
    grid.column['val']['dis'] = 'Value'
    grid.column['val']['unit'] = 'kW'
    grid.extend([
        {
            'siteName': 'Site 1',
            'val': hszinc.Quantity(356.214, 'kW'),
        },
        {
            'siteName': 'Site 2',
            'val': hszinc.Quantity(463.028, 'kW'),
        },
    ])
    return grid


def test_metadata():
    grid = make_metadata_grid()
    grid_str = hszinc.dump(grid)
    assert grid_str == METADATA_EXAMPLE


def test_metadata_json():
    grid = make_metadata_grid()
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == METADATA_EXAMPLE_JSON


def test_multi_grid():
    grids = [make_simple_grid(), make_metadata_grid()]
    grid_str = hszinc.dump(grids)

    assert grid_str == '\n'.join([SIMPLE_EXAMPLE, METADATA_EXAMPLE])


def test_multi_grid_json():
    grids = [make_simple_grid(), make_metadata_grid()]
    grid_json = json.loads(hszinc.dump(grids, mode=hszinc.MODE_JSON))

    assert grid_json[0] == SIMPLE_EXAMPLE_JSON
    assert grid_json[1] == METADATA_EXAMPLE_JSON


def make_grid_meta(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    grid.metadata['aString'] = 'aValue'
    grid.metadata['aNumber'] = 3.14159
    grid.metadata['aNull'] = None
    grid.metadata['aMarker'] = hszinc.MARKER
    grid.metadata['aQuantity'] = hszinc.Quantity(123, 'Hz')
    grid.column['empty'] = {}
    return grid


def test_grid_meta():
    grid_str = hszinc.dump(make_grid_meta())
    assert grid_str == '''ver:"2.0" aString:"aValue" aNumber:3.14159 aNull:N aMarker aQuantity:123Hz
empty
'''


def test_grid_meta_json():
    grid_json = json.loads(hszinc.dump(make_grid_meta(),
                                       mode=hszinc.MODE_JSON))
    assert grid_json == {
        'meta': {
            'ver': '2.0',
            'aString': 's:aValue',
            'aNumber': 'n:3.141590',
            'aNull': None,
            'aMarker': 'm:',
            'aQuantity': 'n:123.000000 Hz',
        },
        'cols': [
            {'name': 'empty'},
        ],
        'rows': [],
    }


def make_col_meta(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    col_meta = hszinc.MetadataObject()
    col_meta['aString'] = 'aValue'
    col_meta['aNumber'] = 3.14159
    col_meta['aNull'] = None
    col_meta['aMarker'] = hszinc.MARKER
    col_meta['aQuantity'] = hszinc.Quantity(123, 'Hz')
    grid.column['empty'] = col_meta
    return grid


def test_col_meta():
    grid_str = hszinc.dump(make_col_meta())
    assert grid_str == '''ver:"2.0"
empty aString:"aValue" aNumber:3.14159 aNull:N aMarker aQuantity:123Hz
'''


def test_col_meta_json():
    grid_json = json.loads(hszinc.dump(make_col_meta(),
                                       mode=hszinc.MODE_JSON))
    assert grid_json == {
        'meta': {
            'ver': '2.0',
        },
        'cols': [
            {'name': 'empty',
             'aString': 's:aValue',
             'aNumber': 'n:3.141590',
             'aNull': None,
             'aMarker': 'm:',
             'aQuantity': 'n:123.000000 Hz',
             },
        ],
        'rows': [],
    }


def test_data_types_v2():
    grid = hszinc.Grid(version=hszinc.VER_2_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'A null value',
            'value': None,
        },
        {
            'comment': 'A marker',
            'value': hszinc.MARKER,
        },
        {
            'comment': 'A "remove" object',
            'value': hszinc.REMOVE,
        },
        {
            'comment': 'A boolean, indicating False',
            'value': False,
        },
        {
            'comment': 'A boolean, indicating True',
            'value': True,
        },
        {
            'comment': 'A reference, without value',
            'value': hszinc.Ref('a-ref'),
        },
        {
            'comment': 'A reference, with value',
            'value': hszinc.Ref('a-ref', 'a value'),
        },
        {
            'comment': 'A binary blob',
            'value': hszinc.Bin('text/plain'),
        },
        {
            'comment': 'A quantity',
            'value': hszinc.Quantity(500, 'miles'),
        },
        {
            'comment': 'A quantity without unit',
            'value': hszinc.Quantity(500, None),
        },
        {
            'comment': 'A coordinate',
            'value': hszinc.Coordinate(-27.4725, 153.003),
        },
        {
            'comment': 'A URI',
            'value': hszinc.Uri(u'http://www.example.com#`unicode:\u1234\u5678`'),
        },
        {
            'comment': 'A string',
            'value': u'This is a test\n' \
                     u'Line two of test\n' \
                     u'\tIndented with "quotes", \\backslashes\\ and ' \
                     u'Unicode characters: \u1234\u5678 and a $ dollar sign',
        },
        {
            'comment': 'A date',
            'value': datetime.date(2016, 1, 13),
        },
        {
            'comment': 'A time',
            'value': datetime.time(7, 51, 43, microsecond=12345),
        },
        {
            'comment': 'A timestamp (non-UTC)',
            'value': pytz.timezone('Europe/Berlin').localize( \
                datetime.datetime(2016, 1, 13, 7, 51, 42, 12345)),
        },
        {
            'comment': 'A timestamp (UTC)',
            'value': pytz.timezone('UTC').localize( \
                datetime.datetime(2016, 1, 13, 7, 51, 42, 12345)),
        },
    ])
    grid_str = hszinc.dump(grid)
    ref_str = '''ver:"2.0"
comment,value
"A null value",N
"A marker",M
"A \\"remove\\" object",R
"A boolean, indicating False",F
"A boolean, indicating True",T
"A reference, without value",@a-ref
"A reference, with value",@a-ref "a value"
"A binary blob",Bin(text/plain)
"A quantity",500miles
"A quantity without unit",500
"A coordinate",C(-27.472500,153.003000)
"A URI",`http://www.example.com#\\`unicode:\\u1234\\u5678\\``
"A string","This is a test\\nLine two of test\\n\\tIndented with \\"quotes\\", \\\\backslashes\\\\ and Unicode characters: \\u1234\\u5678 and a \\$ dollar sign"
"A date",2016-01-13
"A time",07:51:43.012345
"A timestamp (non-UTC)",2016-01-13T07:51:42.012345+01:00 Berlin
"A timestamp (UTC)",2016-01-13T07:51:42.012345+00:00 UTC
'''
    assert grid_str == ref_str


def test_data_types_v3():
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'A NA',
            'value': hszinc.NA,
        },
        {
            'comment': 'An empty list',
            'value': [],
        },
        {
            'comment': 'A null value in a list',
            'value': [None],
        },
        {
            'comment': 'A marker in a list',
            'value': [hszinc.MARKER],
        },
        {
            'comment': 'Booleans',
            'value': [True, False],
        },
        {
            'comment': 'References',
            'value': [hszinc.Ref('a-ref'), hszinc.Ref('a-ref', 'a value')],
        },
        {
            'comment': 'A quantity',
            'value': [hszinc.Quantity(500, 'miles')],
        },
        {
            'comment': 'A XStr',
            'value': [hszinc.XStr("hex", 'deadbeef')],
        },
    ])
    grid_str = hszinc.dump(grid)
    ref_str = '''ver:"3.0"
comment,value
"A NA",NA
"An empty list",[]
"A null value in a list",[N]
"A marker in a list",[M]
"Booleans",[T,F]
"References",[@a-ref,@a-ref "a value"]
"A quantity",[500miles]
"A XStr",[hex("deadbeef")]
'''
    assert grid_str == ref_str


def test_data_types_json_v2():
    grid = hszinc.Grid(version=hszinc.VER_2_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'A null value',
            'value': None,
        },
        {
            'comment': 'A marker',
            'value': hszinc.MARKER,
        },
        {
            'comment': 'A remove (2.0 version)',
            'value': hszinc.REMOVE,
        },
        {
            'comment': 'A boolean, indicating False',
            'value': False,
        },
        {
            'comment': 'A boolean, indicating True',
            'value': True,
        },
        {
            'comment': 'A reference, without value',
            'value': hszinc.Ref('a-ref'),
        },
        {
            'comment': 'A reference, with value',
            'value': hszinc.Ref('a-ref', 'a value'),
        },
        {
            'comment': 'A binary blob',
            'value': hszinc.Bin('text/plain'),
        },
        {
            'comment': 'A quantity',
            'value': hszinc.Quantity(500, 'miles'),
        },
        {
            'comment': 'A quantity without unit',
            'value': hszinc.Quantity(500, None),
        },
        {
            'comment': 'A coordinate',
            'value': hszinc.Coordinate(-27.4725, 153.003),
        },
        {
            'comment': 'A URI',
            'value': hszinc.Uri('http://www.example.com'),
        },
        {
            'comment': 'A string',
            'value': 'This is a test\n' \
                     'Line two of test\n' \
                     '\tIndented with "quotes" and \\backslashes\\',
        },
        {
            'comment': 'A date',
            'value': datetime.date(2016, 1, 13),
        },
        {
            'comment': 'A time',
            'value': datetime.time(7, 51, 43, microsecond=12345),
        },
        {
            'comment': 'A timestamp (non-UTC)',
            'value': pytz.timezone('Europe/Berlin').localize( \
                datetime.datetime(2016, 1, 13, 7, 51, 42, 12345)),
        },
        {
            'comment': 'A timestamp (UTC)',
            'value': pytz.timezone('UTC').localize( \
                datetime.datetime(2016, 1, 13, 7, 51, 42, 12345)),
        },
    ])
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == {
        'meta': {'ver': '2.0'},
        'cols': [
            {'name': 'comment'},
            {'name': 'value'},
        ],
        'rows': [
            {'comment': 's:A null value',
             'value': None},
            {'comment': 's:A marker',
             'value': 'm:'},
            {'comment': 's:A remove (2.0 version)',
             'value': 'x:'},
            {'comment': 's:A boolean, indicating False',
             'value': False},
            {'comment': 's:A boolean, indicating True',
             'value': True},
            {'comment': 's:A reference, without value',
             'value': 'r:a-ref'},
            {'comment': 's:A reference, with value',
             'value': 'r:a-ref a value'},
            {'comment': 's:A binary blob',
             'value': 'b:text/plain'},
            {'comment': 's:A quantity',
             'value': 'n:500.000000 miles'},
            {'comment': 's:A quantity without unit',
             'value': 'n:500.000000'},
            {'comment': 's:A coordinate',
             'value': 'c:-27.472500,153.003000'},
            {'comment': 's:A URI',
             'value': 'u:http://www.example.com'},
            {'comment': 's:A string',
             'value': 's:This is a test\n' \
                      'Line two of test\n' \
                      '\tIndented with \"quotes\" ' \
                      'and \\backslashes\\'},
            {'comment': 's:A date',
             'value': 'd:2016-01-13'},
            {'comment': 's:A time',
             'value': 'h:07:51:43.012345'},
            {'comment': 's:A timestamp (non-UTC)',
             'value': 't:2016-01-13T07:51:42.012345+01:00 Berlin'},
            {'comment': 's:A timestamp (UTC)',
             'value': 't:2016-01-13T07:51:42.012345+00:00 UTC'},
        ],
    }


def test_data_types_json_v3():
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'A Remove (3.0 version)',
            'value': hszinc.REMOVE,
        },
        {
            'comment': 'A NA',
            'value': hszinc.NA,
        },
        {
            'comment': 'An empty list',
            'value': [],
        },
        {
            'comment': 'A null value in a list',
            'value': [None],
        },
        {
            'comment': 'A marker in a list',
            'value': [hszinc.MARKER],
        },
        {
            'comment': 'Booleans',
            'value': [True, False],
        },
        {
            'comment': 'References',
            'value': [hszinc.Ref('a-ref'), hszinc.Ref('a-ref', 'a value')],
        },
        {
            'comment': 'A quantity',
            'value': [hszinc.Quantity(500, 'miles')],
        },
        {
            'comment': 'A XStr',
            'value': [hszinc.XStr("hex", 'deadbeef')],
        },
    ])
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == {
        'meta': {
            'ver': '3.0'
        },
        'cols': [
            {'name': 'comment'},
            {'name': 'value'},
        ],
        'rows': [
            {
                'comment': 's:A Remove (3.0 version)',
                'value': '-:'
            },
            {
                'comment': 's:A NA',
                'value': 'z:'
            },
            {
                'comment': "s:An empty list",
                'value': []
            },
            {
                'comment': "s:A null value in a list",
                'value': [None]
            },
            {
                'comment': "s:A marker in a list",
                'value': ['m:']
            },
            {
                'comment': "s:Booleans",
                'value': [True, False]
            },
            {
                'comment': "s:References",
                'value': ['r:a-ref', 'r:a-ref a value']
            },
            {
                'comment': "s:A quantity",
                'value': ['n:500.000000 miles']  # Python is more precise
                # than The Proclaimers
            },
            {
                'comment': "s:A XStr",
                'value': ['x:hex:deadbeef']
            }
        ]
    }


def test_scalar_dict_json_v3():
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'An empty dict',
            'value': {},
        },
        {
            'comment': 'A marker in a dict',
            'value': {"marker": hszinc.MARKER},
        },
        {
            'comment': 'A references in a dict',
            'value': {"ref": hszinc.Ref('a-ref'), "ref2": hszinc.Ref('a-ref', 'a value')},
        },
        {
            'comment': 'A quantity in a dict',
            'value': {"quantity": hszinc.Quantity(500, 'miles')},
        },
    ])
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == {
        'meta': {
            'ver': '3.0'
        },
        'cols': [
            {'name': 'comment'},
            {'name': 'value'},
        ],
        'rows': [
            {
                'comment': "s:An empty dict",
                'value': {}
            },
            {
                'comment': "s:A marker in a dict",
                'value': {'marker': 'm:'}
            },
            {
                'comment': "s:A references in a dict",
                'value': {'ref': 'r:a-ref', 'ref2': 'r:a-ref a value'}
            },
            {
                'comment': "s:A quantity in a dict",
                'value': {"quantity": 'n:500.000000 miles'}
            }
        ]
    }


def test_scalar_dict_zinc_v3():
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'An empty dict',
            'value': {},
        },
        {
            'comment': 'A marker in a dict',
            'value': {"marker": hszinc.MARKER},
        },
        {
            'comment': 'A references in a dict',
            'value': {"ref": hszinc.Ref('a-ref'), "ref2": hszinc.Ref('a-ref', 'a value')},
        },
        {
            'comment': 'A quantity in a dict',
            'value': {"quantity": hszinc.Quantity(500, 'miles')},
        },
    ])
    grid_str = hszinc.dump(grid, mode=hszinc.MODE_ZINC)
    assert grid_str == ("ver:\"3.0\"\n"
                        "comment,value\n"
                        "\"An empty dict\",{}\n"
                        "\"A marker in a dict\",{marker:M}\n"
                        "\"A references in a dict\",{" +
                        " ".join([str(k) + ":" + str(v) for k, v in {"ref": "@a-ref", "ref2": "@a-ref"}.items()]) \
                        .replace("ref2:@a-ref", "ref2:@a-ref \"a value\"") + \
                        "}\n"
                        "\"A quantity in a dict\",{quantity:500miles}\n")

def test_scalar_dict_zinc_ver():
    # Test that versions are respected.
    try:
        hszinc.dump_scalar({"a":"b"},
                           mode=hszinc.MODE_ZINC, version=hszinc.VER_2_0)
        assert False, 'Serialised a list in Haystack v2.0'
    except ValueError:
        pass


def test_scalar_unknown():
    try:
        hszinc.dump_scalar(hszinc.VER_2_0,
                           mode=hszinc.MODE_ZINC, version=hszinc.VER_2_0)
        assert False, 'Serialised a list in Haystack v2.0'
    except NotImplementedError:
        pass

def test_list_zinc_v2():
    try:
        grid = hszinc.Grid(version=hszinc.VER_2_0)
        grid.column['comment'] = {}
        grid.column['value'] = {}
        grid.extend([
            {
                'comment': 'An empty list',
                'value': [],
            }
        ])
        hszinc.dump(grid, mode=hszinc.MODE_ZINC)
        assert False, 'Project Haystack 2.0 doesn\'t support lists'
    except ValueError:
        pass


def test_dict_zinc_v2():
    try:
        grid = hszinc.Grid(version=hszinc.VER_2_0)
        grid.column['comment'] = {}
        grid.column['value'] = {}
        grid.extend([
            {
                'comment': 'An empty dict',
                'value': {},
            }
        ])
        hszinc.dump(grid, mode=hszinc.MODE_ZINC)
        assert False, 'Project Haystack 2.0 doesn\'t support dict'
    except ValueError:
        pass


def test_list_json_v2():
    try:
        grid = hszinc.Grid(version=hszinc.VER_2_0)
        grid.column['comment'] = {}
        grid.column['value'] = {}
        grid.extend([
            {
                'comment': 'An empty list',
                'value': [],
            }
        ])
        hszinc.dump(grid, mode=hszinc.MODE_JSON)
        assert False, 'Project Haystack 2.0 doesn\'t support lists'
    except ValueError:
        pass


def test_dict_json_v2():
    try:
        grid = hszinc.Grid(version=hszinc.VER_2_0)
        grid.column['comment'] = {}
        grid.column['value'] = {}
        grid.extend([
            {
                'comment': 'An empty dict',
                'value': {},
            }
        ])
        hszinc.dump(grid, mode=hszinc.MODE_JSON)
        assert False, 'Project Haystack 2.0 doesn\'t support dict'
    except ValueError:
        pass


def test_scalar_zinc():
    # No need to be exhaustive, the underlying function is tested heavily by
    # the grid dump tests.
    assert hszinc.dump_scalar(hszinc.Ref('areference', 'a display name'),
                              mode=hszinc.MODE_ZINC) == '@areference "a display name"'


def test_scalar_list_zinc_ver():
    # Test that versions are respected.
    try:
        hszinc.dump_scalar(["a list is not allowed in v2.0"],
                           mode=hszinc.MODE_ZINC, version=hszinc.VER_2_0)
        assert False, 'Serialised a list in Haystack v2.0'
    except ValueError:
        pass


def test_scalar_list_na_ver():
    # Test that versions are respected.
    try:
        hszinc.dump_scalar(hszinc.NA,
                           mode=hszinc.MODE_ZINC, version=hszinc.VER_2_0)
        assert False, 'Serialised a NA in Haystack v2.0'
    except ValueError:
        pass


def test_scalar_json():
    # No need to be exhaustive, the underlying function is tested heavily by
    # the grid dump tests.
    assert hszinc.dump_scalar(hszinc.Ref('areference', 'a display name'),
                              mode=hszinc.MODE_JSON) == 'r:areference a display name'


def test_scalar_list_json_ver():
    # Test that versions are respected.
    try:
        hszinc.dump_scalar(["a list is not allowed in v2.0"],
                           mode=hszinc.MODE_JSON, version=hszinc.VER_2_0)
        assert False, 'Serialised a list in Haystack v2.0'
    except ValueError:
        pass


def test_scalar_na_json_ver():
    # Test that versions are respected.
    try:
        hszinc.dump_scalar(hszinc.NA,
                           mode=hszinc.MODE_JSON, version=hszinc.VER_2_0)
        assert False, 'Serialised a NA in Haystack v2.0'
    except ValueError:
        pass


def test_grid_types_zinc():
    innergrid = hszinc.Grid(version=hszinc.VER_3_0)
    innergrid.column['comment'] = {}
    innergrid.extend([
        {
            'comment': 'A innergrid',
        },
    ])
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['inner'] = {}
    grid.extend([
        {
            'inner': innergrid,
        },
    ])
    grid_str = hszinc.dump(grid, mode=hszinc.MODE_ZINC)
    assert grid_str == ('ver:"3.0"\n'
                        'inner\n'
                        '<<ver:"3.0"\n'
                        'comment\n'
                        '"A innergrid"\n'
                        '>>\n')


def test_grid_types_json():
    innergrid = hszinc.Grid(version=hszinc.VER_3_0)
    innergrid.column['comment'] = {}
    innergrid.extend([
        {
            'comment': 'A innergrid',
        },
    ])
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['inner'] = {}
    grid.extend([
        {
            'inner': innergrid,
        },
    ])
    grid_str = hszinc.dump(grid, mode=hszinc.MODE_JSON)
    grid_json = json.loads(grid_str)
    assert grid_json == {
        'meta': {'ver': '3.0'},
        'cols': [
            {'name': 'inner'},
        ],
        'rows': [
            {'inner': {
                'meta': {'ver': '3.0'},
                'cols': [
                    {'name': 'comment'},
                ],
                'rows': [
                    {'comment': 's:A innergrid'},
                ],
            }},
        ],
    }



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

# If `pint` is not installed or available, skip the relevant tests.
try:
    import pint
    import hszinc
    from hszinc.pintutil import to_pint
    def _enable_pint(pint_en):
        hszinc.use_pint(pint_en)

except ImportError:
    import logging
    logging.warning(
        '`pint` was not available for import.  Tests requiring `pint` will '
        'be skipped.', exc_info=1)

    from nose import SkipTest
    def _enable_pint(pint_en):
        if pint_en:
            raise SkipTest('pint not available')
    def to_pint(*a, **kwa):
        raise SkipTest('pint not available')



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:
import base64
import binascii
import datetime
import random
import string
import sys
import traceback

import six

import hszinc
from hszinc import VER_3_0, Grid, MODE_ZINC, MODE_JSON, XStr
from .pint_enable import to_pint

STR_CHARSET = string.ascii_letters + string.digits + '\n\r\t\f\b'

GENERATION_NUMBER = 1 # FIXME
PERCENT_RECURSIVE = 1


def gen_random_const():
    return random.choice([True, False, None, hszinc.MARKER, hszinc.REMOVE, hszinc.NA])


def gen_random_ref():
    # Generate a randomised reference.
    name = gen_random_str(charset= \
                              string.ascii_letters + string.digits \
                              + '_:-.~')
    if random.choice([True, False]):
        value = gen_random_str(charset= \
                                   string.ascii_letters + string.digits + '_')
    else:
        value = None

    return hszinc.Ref(name, value)


def gen_random_bin():
    # Generate a randomized binary
    return hszinc.Bin(random.choice([
        'text/plain',
        'text/html',
        'text/zinc',
        'application/json',
        'application/octet-stream',
        'image/png',
        'image/jpeg',
    ]))


def gen_random_xstr():
    # Generate a randomized binary
    barray = bytearray(random.getrandbits(8) for _ in range(5))
    return XStr(*random.choice([
        ('hex', binascii.hexlify(barray).decode('ascii')),
        ('b64', binascii.b2a_base64(barray)[:-1] if sys.version_info[0] <= 2
            else binascii.b2a_base64(barray).decode("ascii")
         )
    ]))


def gen_random_uri():
    return hszinc.Uri(gen_random_str(charset= \
                                         string.ascii_letters + string.digits))


def gen_random_str(min_length=1, max_length=20, charset=STR_CHARSET):
    # Generate a random 20-character string
    return ''.join([random.choice(charset) for c in range(0,
                                                          random.randint(min_length, max_length))])


def gen_random_date():
    # This might generate an invalid date, we keep trying until we get one.
    while True:
        try:
            return datetime.date(random.randint(1, 3000),
                                 random.randint(1, 12), random.randint(1, 31))
        except ValueError:
            pass


def gen_random_time():
    return datetime.time(random.randint(0, 23), random.randint(0, 59),
                         random.randint(0, 59), random.randint(0, 999999))


def gen_random_date_time():
    # Pick a random timezone
    tz_name = random.choice(list(hszinc.zoneinfo.get_tz_map().keys()))
    tz = hszinc.zoneinfo.timezone(tz_name)
    return tz.localize(datetime.datetime.combine(
        gen_random_date(), gen_random_time()))


def gen_random_coordinate():
    return hszinc.Coordinate( \
        round(gen_random_num(360) - 180.0, 2),
        round(gen_random_num(360) - 180.0, 2))


def gen_random_num(scale=1000, digits=2):
    return round(random.random() * scale, digits)


def gen_random_quantity():
    return hszinc.Quantity(gen_random_num(),
                           to_pint('percent'))


def gen_random_list():
    return [gen_random_scalar() for x in range(0, random.randint(0, 2))]


def gen_random_map():
    return {gen_random_name(): gen_random_scalar() for x in range(0, random.randint(0, 2))}


RANDOM_TYPES = [
    # Only for v2.0 gen_random_bin,
    gen_random_xstr,
    # gen_random_const, gen_random_ref, gen_random_uri, gen_random_xstr,
    # gen_random_str, gen_random_date, gen_random_time, gen_random_date_time,
    # gen_random_coordinate, gen_random_num, gen_random_quantity
]


def gen_random_scalar():
    if (random.randint(0, 100) < PERCENT_RECURSIVE):
        return random.choice(RANDOM_RECURSIVE_TYPES)()
    else:
        return random.choice(RANDOM_TYPES)()


def gen_random_name(existing=None):
    while True:
        meta = random.choice(string.ascii_lowercase) \
               + gen_random_str(min_length=0, max_length=7, \
                                charset=string.ascii_letters + string.digits)
        if (existing is None) or (meta not in existing):
            return meta


def gen_random_meta():
    meta = hszinc.MetadataObject()
    names = set()
    for n in range(0, random.randint(1, 5)):
        name = gen_random_name(existing=names)
        value = gen_random_scalar()
        meta[name] = value
    return meta


def gen_random_grid():
    # Generate a randomised grid of values and try parsing it back.
    grid = hszinc.Grid(version=VER_3_0)
    grid.metadata.extend(gen_random_meta())

    # Randomised columns
    for n in range(0, random.randint(1, 5)):
        col_name = gen_random_name(existing=grid.column)
        if random.choice([True, False]):
            grid.column[col_name] = gen_random_meta()
        else:
            grid.column[col_name] = {}

    # Randomised rows
    for n in range(0, random.randint(0, 20)):
        row = {}
        for c in grid.column.keys():
            if random.choice([True, False]):
                row[c] = gen_random_scalar()
        grid.append(row)

    return grid


RANDOM_RECURSIVE_TYPES = [gen_random_list, gen_random_map, gen_random_grid]


def dump_grid(g):
    print('Version: %s' % g.version)
    print('Metadata:')
    for k, v in g.metadata.items():
        print('   %s = %r' % (k, v))
    print('Columns:')
    for c, meta in g.column.items():
        print('   %s:' % c)
        for k, v in g.column[c].items():
            print('      %s = %r' % (k, v))
    print('Rows:')
    for row in g:
        print('---')
        for c, v in row.items():
            print('  %s = %r' % (c, v))


def approx_check(v1, v2):
    # Check types match
    if (isinstance(v1, six.string_types) \
            and isinstance(v2, six.string_types)):
        assert type(v1) == type(v2), '%s != %s' % (type(v1), type(v2))
    if isinstance(v1, datetime.time):
        assert v1.replace(microsecond=0) == v2.replace(microsecond=0)
    elif isinstance(v1, datetime.datetime):
        assert v1.tzinfo == v2.tzinfo
        assert v1.date() == v2.date()
        approx_check(v1.time(), v2.time())
    elif isinstance(v1, hszinc.Quantity):
        assert v1.unit == v2.unit
        approx_check(v1.value, v2.value)
    elif isinstance(v1, hszinc.Coordinate):
        approx_check(v1.latitude, v2.latitude)
        approx_check(v1.longitude, v2.longitude)
    elif isinstance(v1, float) or isinstance(v2, float):
        assert abs(v1 - v2) < 0.000001
    elif isinstance(v1, Grid):
        approx_check_grid(v1, v2)
    else:
        assert v1 == v2, '%r != %r' % (v1, v2)


def _try_dump_parse(ref_grid, mode):
    try:
        # Dump the randomised grid to a string
        grid_str = hszinc.dump(ref_grid, mode=mode)
    except:
        # Dump some detail about the grid
        print('Failed to dump grid.')
        dump_grid(ref_grid)
        raise

    # Parse the grid string
    try:
        parsed_grid = hszinc.parse(grid_str, mode=mode, single=True)
    except:
        print('Failed to parse dumped grid')
        dump_grid(ref_grid)
        print('--- Parsed string ---')
        print(grid_str)
        raise

    approx_check_grid(parsed_grid, ref_grid)


def try_dump_parse_json():
    ref_grid = gen_random_grid()
    _try_dump_parse(ref_grid, MODE_JSON)


def try_dump_parse_zinc():
    ref_grid = gen_random_grid()
    _try_dump_parse(ref_grid, MODE_ZINC)


def approx_check_grid(parsed_grid, ref_grid):
    # Check metadata matches
    try:
        assert set(ref_grid.metadata.keys()) \
               == set(parsed_grid.metadata.keys())
        for key in ref_grid.metadata.keys():
            approx_check(ref_grid.metadata[key], parsed_grid.metadata[key])
    except:
        print(traceback.format_exc())
        print('Mismatch in metadata')
        print('Reference grid')
        dump_grid(ref_grid)
        print('Parsed grid')
        dump_grid(parsed_grid)
        raise

    try:
        # Check column matches
        assert set(ref_grid.column.keys()) \
               == set(parsed_grid.column.keys())
    except:
        print(traceback.format_exc())
        print('Mismatch in column')
        print('Reference grid')
        dump_grid(ref_grid)
        print('Parsed grid')
        dump_grid(parsed_grid)
        raise
    for col in ref_grid.column.keys():
        try:
            for key in ref_grid.column[col].keys():
                approx_check(ref_grid.column[col][key], \
                             parsed_grid.column[col][key])
        except:
            print(traceback.format_exc())
            print('Mismatch in metadata for column %s' % col)
            print('Reference: %r' % ref_grid.column[col])
            print('Parsed:    %r' % parsed_grid.column[col])
            raise

    try:
        # Check row matches
        assert len(ref_grid) == len(parsed_grid)
    except:
        print(traceback.format_exc())
        print('Mismatch in row count')
        print('Reference grid')
        dump_grid(ref_grid)
        print('Parsed grid')
        dump_grid(parsed_grid)

    for (ref_row, parsed_row) in zip(ref_grid, parsed_grid):
        try:
            for col in ref_grid.column.keys():
                approx_check(ref_row.get(col), parsed_row.get(col))
        except:
            print(traceback.format_exc())
            print('Mismatch in row')
            print('Reference:')
            print(ref_row)
            print('Parsed:')
            print(parsed_row)
            raise
    assert parsed_grid == ref_grid


def test_loopback_zinc():
    for trial in range(0, GENERATION_NUMBER):
        try_dump_parse_zinc()


def test_loopback_json():
    for trial in range(0, GENERATION_NUMBER):
        try_dump_parse_json()



# (C) 2020 Philippe PRADOS
# -*- coding: utf-8 -*-
# vim: set ts=4 sts=4 et tw=78 sw=4 si:
import datetime
import gc
from datetime import time, date, datetime

from iso8601 import iso8601

from hszinc import Grid, Uri, Ref, Coordinate, MARKER, XStr
from hszinc.filter_ast import FilterUnary, FilterBinary, FilterPath, FilterAST
from hszinc.grid_filter import hs_filter, _FnWrapper, filter_function
from hszinc.zoneinfo import timezone


def test_filter_ast():
    assert repr(FilterAST(None)) == 'AST:None'


def test_filter_tag_only():
    result = hs_filter.parseString('geo', parseAll=True)[0]
    assert isinstance(result, FilterUnary)
    assert result.op == "has"


def test_filter_tag_equal_values():
    result = hs_filter.parseString('bool == true', parseAll=True)[0]  # Zinc use TF
    assert isinstance(result.right, bool)
    assert result.right is True

    result = hs_filter.parseString('number == 1', parseAll=True)[0]
    assert isinstance(result.right, float)
    assert result.right == 1.0

    result = hs_filter.parseString('number == 1.0', parseAll=True)[0]
    assert isinstance(result.right, float)
    assert result.right == 1.0

    result = hs_filter.parseString('str == "str"', parseAll=True)[0]
    assert isinstance(result.right, str)
    assert result.right == "str"

    result = hs_filter.parseString('uri == `uri`', parseAll=True)[0]
    assert isinstance(result.right, Uri)
    assert result.right == Uri("uri")

    result = hs_filter.parseString('date == 1977-04-22', parseAll=True)[0]
    assert isinstance(result.right, date)
    assert result.right == date(1977, 4, 22)

    result = hs_filter.parseString('time == 11:11:11', parseAll=True)[0]
    assert isinstance(result.right, time)
    assert result.right == time(11, 11, 11)

    result = hs_filter.parseString('date_time == 1977-04-22T01:00:00-05:00', parseAll=True)[0]
    assert isinstance(result.right, datetime)
    assert result.right == iso8601.parse_date("1977-04-22T01:00:00-05:00")

    result = hs_filter.parseString('date_time == 1977-04-22T01:00:00 GMT+1', parseAll=True)[0]
    assert isinstance(result.right, datetime)
    assert result.right == iso8601.parse_date("1977-04-22T01:00:00").astimezone(timezone("GMT+1"))

    result = hs_filter.parseString('date_time == 1977-04-22T01:00:00 Paris', parseAll=True)[0]
    assert isinstance(result.right, datetime)
    assert result.right == iso8601.parse_date("1977-04-22T01:00:00").astimezone(timezone("Paris"))

    # --- Extended syntax (types not in specification)
    result = hs_filter.parseString('ref == @abc', parseAll=True)[0]
    assert isinstance(result.right, Ref)
    assert result.right == Ref("abc")

    result = hs_filter.parseString('geo == @abc "a"', parseAll=True)[0]
    assert isinstance(result.right, Ref)
    assert result.right == Ref("abc", "a")

    result = hs_filter.parseString('ref == C(1.0,-1.0)', parseAll=True)[0]
    assert isinstance(result.right, Coordinate)
    assert result.right == Coordinate(1.0, -1.0)

    result = hs_filter.parseString('list == [ 1, 2 ]', parseAll=True)[0]
    assert isinstance(result.right, list)
    assert result.right == [1, 2]

    result = hs_filter.parseString('dict == { a b:2 }', parseAll=True)[0]
    assert isinstance(result.right, dict)
    assert result.right == {"a": MARKER, "b": 2.0}

    result = hs_filter.parseString('map == hex("010203")', parseAll=True)[0]
    assert isinstance(result.right, XStr)
    assert result.right == XStr("hex", "010203")


def test_filter_tag_comparaison_operator():
    result = hs_filter.parseString('bool == true', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == '=='

    result = hs_filter.parseString('bool <= true', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == '<='

    result = hs_filter.parseString('bool >= true', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == '>='

    result = hs_filter.parseString('bool != true', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == '!='

    result = hs_filter.parseString('bool < true', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == '<'

    result = hs_filter.parseString('bool > true', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == '>'


def test_filter_navigation():
    result = hs_filter.parseString('a->b', parseAll=True)[0]
    assert isinstance(result, FilterUnary)
    assert isinstance(result.right, FilterPath)
    assert result.right.path == ['a', 'b']

    result = hs_filter.parseString('a->b->c', parseAll=True)[0]
    assert isinstance(result, FilterUnary)
    assert isinstance(result.right, FilterPath)
    assert result.right.path == ['a', 'b', 'c']


def test_filter_boolean_operator():
    result = hs_filter.parseString('bool and bool', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == "and"

    result = hs_filter.parseString('bool==true and bool!=false', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == "and"
    assert result.left.op == "=="
    assert result.right.op == "!="

    result = hs_filter.parseString('bool>=true or bool<=false', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == "or"
    assert result.left.op == ">="
    assert result.right.op == "<="

    result = hs_filter.parseString('(bool<true or bool>false) and not bool', parseAll=True)[0]
    assert isinstance(result, FilterBinary)
    assert result.op == "and"
    assert isinstance(result.left, FilterBinary)
    assert result.left.op == "or"
    assert isinstance(result.right, FilterUnary)
    assert result.right.op == 'not'

    result = hs_filter.parseString('equip and siteRef->geoCity == "Chicago"', parseAll=True)[0]
    repr(result)
    assert isinstance(result, FilterBinary)
    assert result.op == "and"
    assert isinstance(result.left, FilterUnary)
    assert result.left.op == "has"
    assert isinstance(result.right, FilterBinary)
    assert result.right.op == "=="
    assert result.right.left.path == ["siteRef", "geoCity"]


def test_generated_filter():
    assert filter_function('equip == "Chicago"')(None, {"equip": "Chicago"})
    assert filter_function('equip == "Chicago" or not titi')(None, {"equip": "Chicago"})
    assert filter_function('equip == "Chicago" or not titi')(None, {"equip": "NewYork"})
    assert not filter_function('equip == "Chicago" or not titi')(None, {"titi": MARKER})


def test_generated_filter_with_reference():
    grid = Grid(columns={'id': {}})
    grid.insert(0, {'id': 'id1'})
    assert filter_function('ref == @id1')(None, {"ref": Ref("id1")})


def test_grid_filter():
    grid = Grid(columns={'id': {}, 'site': {}, 'equip':{},'geoPostalCode':{},'ahu':{},
                         'geoCity':{},'curVal':{},'hvac':{},'siteRef':{}})
    grid.append({'id': 'id1', 'site': MARKER, 'equip': 'Chicago', 'geoPostalCode': "78280", 'ahu': MARKER,
                 'geoCity': 'Chicago', 'curVal': 76})
    grid.append({'id': 'id2', 'hvac': MARKER, 'geoPostalCode': "23220", 'curVal': 75})
    grid.append({'equip': 'Chicago', 'hvac': MARKER, 'siteRef': Ref('id1'), 'curVal': 74})
    result = grid.filter('site')
    assert len(result) == 1
    assert result['id1']

    result = grid.filter('equip == "Chicago"')
    assert len(result) == 2
    assert result[0]['id'] == 'id1'
    assert result[1]['equip'] == 'Chicago'

    result = grid.filter('not id')
    assert len(result) == 1
    assert result[0]['equip'] == 'Chicago'


def test_grid_specification_filter_sample():
    grid = Grid(columns={'id': {}, 'site': {}, 'equip':{},'geoPostalCode':{},'ahu':{},
                         'geoCity':{},'curVal':{},'hvac':{},'siteRef':{}})
    grid.append({'id': 'id1', 'site': MARKER, 'equip': 'Chicago', 'geoPostalCode': "78280", 'ahu': MARKER,
                 'geoCity': 'Chicago', 'curVal': 76})
    grid.append({'id': 'id2', 'hvac': MARKER, 'geoPostalCode': "23220", 'curVal': 75})
    grid.append({'equip': 'Chicago', 'hvac': MARKER, 'siteRef': Ref('id1'), 'curVal': 74})

    # specification samples
    result = grid.filter('geoPostalCode == "23220"')
    assert len(result) == 1
    assert result[0]['id'] == 'id2'

    result = grid.filter('geoPostalCode != "23220"')
    assert len(result) == 1
    assert result[0]['id'] == 'id1'

    result = grid.filter('curVal < 75')
    assert len(result) == 1
    assert result[0]['siteRef'] == Ref('id1')

    result = grid.filter('curVal <= 75')
    assert len(result) == 2
    assert result[0]['id'] == 'id2'
    assert result[1]['siteRef'] == Ref('id1')

    result = grid.filter('curVal > 75')
    assert len(result) == 1
    assert result[0]['id'] == 'id1'

    result = grid.filter('curVal >= 75')
    assert len(result) == 2
    assert result[0]['id'] == 'id1'
    assert result[1]['id'] == 'id2'

    result = grid.filter('site or equip')
    assert len(result) == 2
    assert result[0]['id'] == 'id1'
    assert result[1]['siteRef'] == Ref('id1')

    result = grid.filter('equip and hvac')
    assert len(result) == 1
    assert result[0]['siteRef'] == Ref('id1')

    result = grid.filter('equip and not ahu')
    assert len(result) == 1
    assert result[0]['siteRef'] == Ref('id1')

    result = grid.filter('equip and siteRef->geoCity == "Chicago"')
    assert len(result) == 1
    assert result[0]['equip'] == 'Chicago'


def test_if_generated_function_removed():
    # Check if the generated function will be removed
    import hszinc.grid_filter
    wrapper = _FnWrapper("_acme", "def _acme(): pass")
    assert hszinc.grid_filter._acme
    del wrapper
    gc.collect()
    try:
        hszinc.grid_filter._acme
        assert False
    except AttributeError:
        pass


def test_slide_get():
    grid = Grid(columns={'id': {}, 'site': {}, 'equip':{},'geoPostalCode':{},'ahu':{},
                         'geoCity':{},'curVal':{},'hvac':{},'siteRef':{}})
    grid.append({'id': 'id1', 'site': MARKER, 'equip': 'Chicago', 'geoPostalCode': "78280", 'ahu': MARKER,
                 'geoCity': 'Chicago', 'curVal': 76})
    grid.append({'id': 'id2', 'hvac': MARKER, 'geoPostalCode': "23220", 'curVal': 75})
    grid.append({'equip': 'Chicago', 'hvac': MARKER, 'siteRef': Ref('id1'), 'curVal': 74})
    assert len(grid[0:1]) == 1
    assert len(grid[1:]) == 2
    assert len(grid[:]) == 3


def test_empty_filter():
    grid = Grid(columns={'id': {}, 'site': {}, 'equip':{},'geoPostalCode':{},'ahu':{},
                         'geoCity':{},'curVal':{},'hvac':{},'siteRef':{}})
    grid.append({'id': 'id1', 'site': MARKER, 'equip': 'Chicago', 'geoPostalCode': "78280", 'ahu': MARKER,
                 'geoCity': 'Chicago', 'curVal': 76})
    grid.append({'id': 'id2', 'hvac': MARKER, 'geoPostalCode': "23220", 'curVal': 75})
    grid.append({'equip': 'Chicago', 'hvac': MARKER, 'siteRef': Ref('id1'), 'curVal': 74})
    assert len(grid.filter('')) == 3


def test_empty_filter_and_limit():
    grid = Grid(columns={'id': {}, 'site': {}})
    grid.append({'id': 'id1', 'site': MARKER, 'equip': 'Chicago', 'geoPostalCode': "78280", 'ahu': MARKER,
                 'geoCity': 'Chicago', 'curVal': 76})
    grid.append({'id': 'id2', 'hvac': MARKER, 'geoPostalCode': "23220", 'curVal': 75})
    grid.append({'equip': 'Chicago', 'hvac': MARKER, 'siteRef': Ref('id1'), 'curVal': 74})
    assert len(grid.filter('', limit=1)) == 1
    assert len(grid.filter('', limit=2)) == 2


def test_filter_and_limit():
    grid = Grid(columns={'id': {}, 'site': {}})
    grid.append({'id': 'id1', 'site': MARKER, 'equip': 'Chicago', 'geoPostalCode': "78280", 'ahu': MARKER,
                 'geoCity': 'Chicago', 'curVal': 76})
    grid.append({'id': 'id2', 'hvac': MARKER, 'geoPostalCode': "23220", 'curVal': 75})
    grid.append({'equip': 'Chicago', 'hvac': MARKER, 'siteRef': Ref('id1'), 'curVal': 74})

    assert len(grid.filter('not acme', limit=1)) == 1



from datetime import datetime
try:
    from functools import lru_cache
except ImportError:  # pragma: no cover
    from backports.functools_lru_cache import lru_cache

from iso8601 import iso8601
from pyparsing import Word, ZeroOrMore, Literal, Forward, Combine, Optional, Regex, OneOrMore, \
    CaselessLiteral, Suppress, Group

from .datatypes import *
from .filter_ast import *
from .zincparser import DelimitedList, to_dict
from .zoneinfo import timezone

# TODO: check escape char. Voir hs_str du parser

hs_filter = Forward()
hs_strChar = Regex(r"([^\x00-\x1f\\\"]|\\[bfnrt\\\"$]|\\[uU][0-9a-fA-F]{4})")
hs_str = Combine(Suppress(Literal('"')) + ZeroOrMore(hs_strChar) + Suppress(Literal('"')))
hs_uriChar = Regex(r"([^\x00-\x1f\\`]|\\[bfnrt\\:/?" \
                   + r"#\[\]@&=;`]|\\[uU][0-9a-fA-F]{4})")
hs_uri = Combine(Suppress(Literal('`')) + ZeroOrMore(hs_uriChar) + Suppress(Literal('`'))).setParseAction(
    lambda toks: Uri(toks[0])
)
hs_digits = Regex(r'[0-9_]+')
hs_alpha = Regex(r'[a-zA-Z]')
hs_valueSep = Regex(r' *, *')
hs_plusMinus = Literal('+') | Literal('-')
hs_exp = Combine(CaselessLiteral('e') + Optional(hs_plusMinus) + hs_digits)
hs_decimal = Combine(
    Optional(Literal('-')) + hs_digits + Optional(Literal('.') + hs_digits) + Optional(hs_exp)).setParseAction(
    lambda toks: float(toks[0])
)
hs_unitChar = hs_alpha | Word(u'%_/$' + u''.join([
    six.unichr(c)
    for c in range(0x0080, 0xffff)
]), exact=1)
hs_unit = Combine(OneOrMore(hs_unitChar))
hs_digit = Regex(r'\d')
hs_digits = Regex(r'[0-9_]+')
hs_quantity = (hs_decimal + hs_unit).leaveWhitespace().setParseAction(
    lambda toks: Quantity(toks[0], toks[1])
)
hs_number = hs_quantity | hs_decimal | Literal('INF') | Literal("-INF") | Literal("Nan")
hs_bool = (Literal("true") | Literal("false")).setParseAction(
    lambda toks: toks[0] == "true"
)  # Extension to accept T or F
hs_id = Regex(r'[a-z][a-zA-Z0-9_]*')
hs_name = hs_id

hs_coordDeg = Combine(
    Optional(Literal('-')) +
    Optional(Regex(r'[0-9]+')) +
    Optional(Literal('.') + Regex(r'[0-9]+'))).setParseAction(
    lambda toks: [float(toks[0])]
)

hs_coord = (Suppress(Literal('C(')) + \
            hs_coordDeg + \
            Suppress(hs_valueSep) + \
            hs_coordDeg + \
            Suppress(Literal(')'))).setParseAction(
    lambda toks: Coordinate(toks[0], toks[1])
)

# Singleton values
hs_remove = Literal('R').setParseAction( \
    lambda toks: [REMOVE]).setName('remove')
hs_marker = Literal('M').setParseAction( \
    lambda toks: [MARKER]).setName('marker')
hs_null = Literal('N').setParseAction( \
    lambda toks: [None]).setName('null')
hs_na = Literal('NA').setParseAction( \
    lambda toks: [NA]).setName('na')

hs_binChar = Regex(r"[\x20-\x27\x2a-\x7f]")
hs_bin = Combine(
    Suppress(Literal('Bin(')) +
    Combine(ZeroOrMore(hs_binChar)) +
    Suppress(Literal(')'))
).setParseAction(
    lambda toks: [Bin(toks[0])]
)

hs_xstr = (Regex(r"[a-zA-Z0-9_]+") +
           Suppress(Literal('(')) +
           hs_str +
           Suppress(Literal(')'))).setParseAction(
    lambda toks: [XStr(toks[0], toks[1])]
)

hs_dateSep = CaselessLiteral('T')
hs_date_str = Combine(
    hs_digit + hs_digit + hs_digit + hs_digit +
    Literal('-') +
    hs_digit + hs_digit +
    Literal('-') +
    hs_digit + hs_digit)
hs_date = hs_date_str.copy().setParseAction(
    lambda toks: [datetime.strptime(toks[0], '%Y-%m-%d').date()])

hs_time_str = Combine(
    hs_digit + hs_digit +
    Literal(':') +
    hs_digit + hs_digit +
    Literal(':') +
    hs_digit + hs_digit +
    Optional(
        Literal('.') +
        OneOrMore(hs_digit)))


def _parse_time(toks):
    time_str = toks[0]
    time_fmt = '%H:%M:%S'
    if '.' in time_str:
        time_fmt += '.%f'
    return [datetime.strptime(time_str, time_fmt).time()]


hs_time = hs_time_str.copy().setParseAction(_parse_time)

hs_tzHHMMOffset = Combine(
    CaselessLiteral('z') |
    (hs_plusMinus + Regex(r'\d\d:\d\d')))

hs_isoDateTime = Combine(
    hs_date_str +
    hs_dateSep +
    hs_time_str +
    Optional(hs_tzHHMMOffset)).setParseAction(
    lambda toks: [iso8601.parse_date(toks[0].upper())]
)

hs_tzName = Regex(r'[A-Z][a-zA-Z0-9_\-]*')
hs_tzUTCGMT = Literal('UTC') | Literal('GMT')
hs_tzUTCOffset = Combine(
    hs_tzUTCGMT + Optional(
        Literal('0') | (hs_plusMinus + OneOrMore(hs_digit))))
hs_timeZoneName = hs_tzUTCOffset | hs_tzName


def _parse_datetime(toks):
    # Made up of parts: ISO8601 Date/Time, time zone label
    isodt = toks[0]
    if len(toks) > 1:
        tzname = toks[1]
    else:
        tzname = None

    if (isodt.tzinfo is None) and bool(tzname):  # pragma: no cover
        # This technically shouldn't happen according to Zinc specs
        return [timezone(tzname).localise(isodt)]
    elif bool(tzname):
        try:
            tz = timezone(tzname)
            return [isodt.astimezone(tz)]
        except:  # pragma: no cover
            # Unlikely to occur, might do though if Project Haystack changes
            # its timezone list or if a system doesn't recognise a particular
            # timezone.
            return [isodt]  # Failed, leave alone
    else:
        return [isodt]


hs_dateTime = (hs_isoDateTime + \
               Optional(
                   hs_timeZoneName
               )).setParseAction(_parse_datetime)

hs_val = Forward()
hs_list = Group( \
    Suppress(Regex(r'[ *]')) |
    (Suppress(Regex(r'\[ *')) +
     Optional(DelimitedList( \
         hs_val, \
         delim=hs_valueSep)) + \
     Suppress(Regex(r' *\]')) \
     )
).setParseAction(lambda toks: toks.asList())

hs_tagmarker = hs_id
hs_tagpair = hs_id + Literal(':') + hs_val
hs_tag = hs_tagpair | hs_tagmarker

hs_tags = ZeroOrMore(hs_tag)

hs_dict = (Suppress(Literal('{')) + \
           hs_tags + \
           Suppress(Literal('}'))).setParseAction(to_dict)

hs_refChar = hs_alpha | hs_digit | Word('_:-.~', exact=1)
hs_ref = (Suppress(Literal('@')) + Combine(ZeroOrMore(hs_refChar)) + Optional(
    hs_str)).setParseAction(
    lambda toks: [Ref(toks[0], toks[1] if len(toks) > 1 else None)])

hs_val <<= hs_list | hs_dict | \
           hs_ref | hs_bin | hs_xstr | \
           hs_dateTime | hs_date | hs_time | \
           hs_coord | \
           hs_number | hs_na | hs_null | hs_marker | hs_bool | \
           hs_str | hs_uri

hs_path = (hs_name + ZeroOrMore(Suppress(Literal("->")) + hs_name)).setParseAction(
    lambda toks: FilterPath([t for t in toks])
)
hs_cmpOp = Literal("==") | Literal("!=") | Literal("<=") | Literal(">=") | Literal("<") | Literal(">")
hs_cmp = (hs_path + hs_cmpOp + hs_val).setParseAction(
    lambda toks: FilterBinary(toks[1], toks[0], toks[2])
)
hs_missing = (Suppress(Literal("not")) + hs_path).setParseAction(
    lambda toks: FilterUnary("not", toks[0])
)
hs_has = hs_path.copy().setParseAction(
    lambda toks: FilterUnary("has", FilterPath([t for t in toks]))
)

hs_parens = (Suppress(Literal("(")) + hs_filter + Suppress(Literal(")"))).setParseAction(
    lambda toks: toks[0]
)
hs_term = hs_parens | hs_missing | hs_cmp | hs_has
hs_condAnd = (hs_term + ZeroOrMore(Literal("and") + hs_term)).setParseAction(
    lambda toks: FilterBinary("and", toks[0], toks[2]) if len(toks) > 1 else toks[0]
)
hs_condOr = (hs_condAnd + ZeroOrMore(Literal("or") + hs_condAnd)).setParseAction(
    lambda toks: FilterBinary("or", toks[0], toks[2]) if len(toks) > 1 else toks[0]
)
hs_filter <<= hs_condOr


def parse_filter(filter):
    '''
    Return an AST tree of filter.
    Can be used to generate other language (SQL, etc.)
    '''
    return FilterAST(hs_filter.parseString(filter, parseAll=True)[0])


## --- Generate python to apply filter
FILTER_CACHE_LRU_SIZE = 500
_id_function = 0


class _NotFoundValue():
    def __repr__(self):
        return 'NOT_FOUND'

    def __bool__(self):
        return False

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return False


NOT_FOUND = _NotFoundValue()


def _get_path(grid, obj, paths):
    try:
        for i, path in enumerate(paths):
            obj = obj[path]
            if i != len(paths)-1 and isinstance(obj, Ref):
                obj = grid[obj.name]  # Follow the reference
        return obj  # It's a value at this time
    except KeyError:
        return NOT_FOUND


def _generate_filter_in_python(node, def_filter):
    if isinstance(node, FilterPath):
        def_filter.append("_get_path(_grid, _entity, %s)" % node.path)
    elif isinstance(node, FilterBinary):
        def_filter.append("(")
        def_filter.extend(_generate_filter_in_python(node.left, []))
        def_filter.append(" " + node.op + " ")
        def_filter.extend(_generate_filter_in_python(node.right, []))
        def_filter.append(")")
    elif isinstance(node, FilterUnary):
        if node.op == "has":
            def_filter.append('(id(')
            def_filter.extend(_generate_filter_in_python(node.right, []))
            def_filter.append(') !=  id(NOT_FOUND))')
        elif node.op == "not":
            def_filter.append('(id(')
            def_filter.extend(_generate_filter_in_python(node.right, []))
            def_filter.append(") == id(NOT_FOUND))")
        else:  # pragma: no cover
            assert 0
    else:
        def_filter.append(repr(node))
    return def_filter


class _FnWrapper():
    def __init__(self, fun_name, function_template):
        self.fun_name = fun_name
        exec(function_template, globals(), globals())

    def __del__(self):  # pragma: no cover
        del globals()[self.fun_name]  # Remove generated function if the LRU ask that

    def get(self):
        return globals()[self.fun_name]

@lru_cache(maxsize=FILTER_CACHE_LRU_SIZE)
def _filter_function(filter):
    global _id_function
    def_filter = _generate_filter_in_python(parse_filter(filter)._head, [])
    fun_name = "_gen_hsfilter_" + str(_id_function)
    function_template = "def %s(_grid, _entity):\n  return " % fun_name + "".join(def_filter)
    print("\nGenerate:\n# " + filter + "\n" + function_template)  # FIXME: debug
    _id_function += 1
    return _FnWrapper(fun_name, function_template)


def filter_function(filter):
    return _filter_function(filter).get()



# (C) 2020 Philippe PRADOS
# -*- coding: utf-8 -*-
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

class FilterPath:
    def __init__(self, path):
        self.path = path

    def __repr__(self):
        return "->".join(self.path)


class FilterBinary():
    def __init__(self, op, left, right):
        self.op = op
        self.left = left
        self.right = right

    def __repr__(self):
        return "%s %s %s" % (self.left, self.op, self.right)


class FilterUnary():
    def __init__(self, op, right):
        self.op = op
        self.right = right

    def __repr__(self):
        return "%s %s" % (self.op, self.right)


class FilterAST:
    def __init__(self, head):
        self._head = head

    def __repr__(self):
        return "AST:" + repr(self._head)





#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc grammar specification.
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:
import datetime
import logging
import re
import sys

import iso8601
import pyparsing as pp
import six

# Bring in special Project Haystack types and time zones
from .datatypes import Quantity, Coordinate, Uri, Bin, MARKER, NA, REMOVE, Ref, XStr
from .grid import Grid
# Bring in our sortable dict class to preserve order
from .sortabledict import SortableDict
# Bring in version handling
from .version import Version, VER_2_0, VER_3_0
from .zoneinfo import timezone

# Logging instance for reporting debug info
LOG = logging.getLogger(__name__)

# All grids start with the version string.
VERSION_RE = re.compile(r'^ver:"(([^"\\]|\\[\\"bfnrt$])+)"')
NEWLINE_RE = re.compile(r'\r?\n')

# Character number regex; for exceptions
CHAR_NUM_RE = re.compile(' *\(at char \d+\),')


def reformat_exception(ex_msg, line_num=None):
    print(ex_msg)
    msg = CHAR_NUM_RE.sub(u'', six.text_type(ex_msg))
    print(msg)
    if line_num is not None:
        return msg.replace(u'line:1', u'line:%d' % line_num)
    else:
        return msg


# Convenience function, we want whitespace left alone.
def _leave_ws(cls, *args, **kwargs):
    return cls(*args, **kwargs).leaveWhitespace()


# Versions of the pyparsing types that leave our whitespace alone!
Empty = lambda *a, **kwa: _leave_ws(pp.Empty, *a, **kwa)
Regex = lambda *a, **kwa: _leave_ws(pp.Regex, *a, **kwa)
Literal = lambda *a, **kwa: _leave_ws(pp.Literal, *a, **kwa)
CaselessLiteral = lambda *a, **kwa: _leave_ws(pp.CaselessLiteral, *a, **kwa)
Word = lambda *a, **kwa: _leave_ws(pp.Word, *a, **kwa)
Optional = lambda *a, **kwa: _leave_ws(pp.Optional, *a, **kwa)
Suppress = lambda *a, **kwa: _leave_ws(pp.Suppress, *a, **kwa)
Combine = lambda *a, **kwa: _leave_ws(pp.Combine, *a, **kwa)
And = lambda *a, **kwa: _leave_ws(pp.And, *a, **kwa)
Or = lambda *a, **kwa: _leave_ws(pp.Or, *a, **kwa)
ZeroOrMore = lambda *a, **kwa: _leave_ws(pp.ZeroOrMore, *a, **kwa)
OneOrMore = lambda *a, **kwa: _leave_ws(pp.OneOrMore, *a, **kwa)
Group = lambda *a, **kwa: _leave_ws(pp.Group, *a, **kwa)
DelimitedList = lambda *a, **kwa: _leave_ws(pp.delimitedList, *a, **kwa)
Forward = lambda *a, **kwa: _leave_ws(pp.Forward, *a, **kwa)


class ZincParseException(ValueError):
    """
    Exception thrown when a grid cannot be parsed successfully.  If known,
    the line and column for the grid are given.
    """

    def __init__(self, message, grid_str, line, col):
        self.grid_str = grid_str
        self.line = line
        self.col = col

        try:
            # If we know the line and column, point it out in the message.
            grid_str_lines = grid_str.split('\n')
            width = max([len(l) for l in grid_str_lines])
            linefmt = u'%%-%ds' % width
            rowfmt = u'%4d%s' + linefmt + u'%s'

            formatted_lines = [
                rowfmt % (
                    num,
                    ' >' if (line == num) else '| ',
                    line_str,
                    '< ' if (line == num) else ' |'
                )
                for (num, line_str)
                in enumerate(grid_str.split('\n'), 1)
            ]
            formatted_lines.insert(line,
                                   (u'    | ' + linefmt + u' |') \
                                   % (((col - 2) * u' ') + '.^.')
                                   )

            # Border it for readability
            formatted_lines.insert(0, u'    .' + (u'-' * (2 + width)) + u'.')
            formatted_lines.append(u'    \'' + (u'-' * (2 + width)) + u'\'')

            # Append to message
            message += u'\n%s' % u'\n'.join(formatted_lines)
        except:  # pragma: no cover
            # We should not get here.
            LOG.exception('Exception encountered formatting log message')
            pass

        super(ZincParseException, self).__init__(message)


class NearestMatch(object):
    """
    This class returns the nearest matching grammar for the given version.
    """

    def __init__(self, known_grammars):
        self._known_grammars = known_grammars

    def __getitem__(self, ver):
        """
        Retrieve the grammar that closest matches the version string given.
        """
        try:
            return self._known_grammars[ver]
        except KeyError:
            pass

        nearest = Version.nearest(ver)
        g = self._known_grammars[nearest]
        self._known_grammars[ver] = g
        return g


class GenerateMatch(object):
    """
    This class tries to generate a matching grammar based on the version input given.
    """

    def __init__(self, generator_fn):
        self._generator_fn = generator_fn
        self._known_grammars = {}

    def __getitem__(self, ver):
        try:
            return self._known_grammars[ver]
        except KeyError:
            g = self._generator_fn(ver)
            self._known_grammars[ver] = g
            return g


def _unescape(s, uri=False):
    """
    Iterative parser for string escapes.
    """
    out = ''
    while len(s) > 0:
        c = s[0]
        if c == '\\':
            # Backslash escape
            esc_c = s[1]

            if esc_c in ('u', 'U'):
                # Unicode escape
                out += six.unichr(int(s[2:6], base=16))
                s = s[6:]
                continue
            else:
                if esc_c == 'b':
                    out += '\b'
                elif esc_c == 'f':
                    out += '\f'
                elif esc_c == 'n':
                    out += '\n'
                elif esc_c == 'r':
                    out += '\r'
                elif esc_c == 't':
                    out += '\t'
                else:
                    if uri and (esc_c == '#'):
                        # \# is passed through with backslash.
                        out += '\\'
                    # Pass through
                    out += esc_c
                s = s[2:]
                continue
        else:
            out += c
            s = s[1:]
    return out


# Grammar according to
#   latest: http://project-haystack.org/doc/Zinc
#   "2.0":  https://web.archive.org/web/20141012013653/http://project-haystack.org:80/doc/Zinc
#   "3.0":  https://web.archive.org/web/20160805064015/http://project-haystack.org:80/doc/Zinc

# Rudimentary elements
hs_digit = Regex(r'\d')
hs_digits = Regex(r'[0-9_]+').setParseAction(
    lambda toks: [''.join([t.replace('_', '') for t in toks[0]])])
hs_alphaLo = Regex(r'[a-z]')
hs_alphaHi = Regex(r'[A-Z]')
hs_alpha = Regex(r'[a-zA-Z]')
hs_valueSep = Regex(r' *, *').setName('valueSep')
hs_rowSep = Regex(r' *\n *').setName('rowSep')
hs_plusMinus = Or([Literal('+'), Literal('-')])

# Forward declaration of data types.
hs_scalar_2_0 = Forward()
hs_scalar_3_0 = Forward()
hs_scalar = NearestMatch({
    VER_2_0: hs_scalar_2_0,
    VER_3_0: hs_scalar_3_0
})

hs_grid_2_0 = Forward()
hs_grid_3_0 = Forward()
hs_grid = NearestMatch({
    VER_2_0: hs_grid_2_0,
    VER_3_0: hs_grid_3_0
})

# Co-ordinates
hs_coordDeg = Combine(And([
    Optional(Literal('-')),
    Optional(hs_digits),
    Optional(And([Literal('.'), hs_digits]))
])).setParseAction(lambda toks: [float(toks[0] or '0')])
hs_coord = And([Suppress(Literal('C(')),
                hs_coordDeg,
                Suppress(hs_valueSep),
                hs_coordDeg,
                Suppress(Literal(')'))]).setParseAction(
    lambda toks: [Coordinate(toks[0], toks[1])])

# Dates and times
hs_tzHHMMOffset = Combine(Or([
    CaselessLiteral('z'),
    And([hs_plusMinus, Regex(r'\d\d:\d\d')])]
))
hs_tzName = Regex(r'[A-Z][a-zA-Z0-9_\-]*')
hs_tzUTCGMT = Or([Literal('UTC'), Literal('GMT')])
hs_tzUTCOffset = Combine(And([
    hs_tzUTCGMT, Optional(
        Or([Literal('0'),
            And([hs_plusMinus, OneOrMore(hs_digit)]
                )]
           ))]))
hs_timeZoneName = Or([hs_tzUTCOffset, hs_tzName])
hs_dateSep = CaselessLiteral('T')
hs_date_str = Combine(And([
    hs_digit, hs_digit, hs_digit, hs_digit,
    Literal('-'),
    hs_digit, hs_digit,
    Literal('-'),
    hs_digit, hs_digit]))
hs_date = hs_date_str.copy().setParseAction(
    lambda toks: [datetime.datetime.strptime(toks[0], '%Y-%m-%d').date()])

hs_time_str = Combine(And([
    hs_digit, hs_digit,
    Literal(':'),
    hs_digit, hs_digit,
    Literal(':'),
    hs_digit, hs_digit,
    Optional(And([
        Literal('.'),
        OneOrMore(hs_digit)]))
]))


def _parse_time(toks):
    time_str = toks[0]
    time_fmt = '%H:%M:%S'
    if '.' in time_str:
        time_fmt += '.%f'
    return [datetime.datetime.strptime(time_str, time_fmt).time()]


hs_time = hs_time_str.copy().setParseAction(_parse_time)
hs_isoDateTime = Combine(And([
    hs_date_str,
    hs_dateSep,
    hs_time_str,
    Optional(hs_tzHHMMOffset)
])).setParseAction(lambda toks: [iso8601.parse_date(toks[0].upper())])


def _parse_datetime(toks):
    # Made up of parts: ISO8601 Date/Time, time zone label
    isodt = toks[0]
    if len(toks) > 1:
        tzname = toks[1]
    else:
        tzname = None

    if (isodt.tzinfo is None) and bool(tzname):  # pragma: no cover
        # This technically shouldn't happen according to Zinc specs
        return [timezone(tzname).localise(isodt)]
    elif bool(tzname):
        try:
            tz = timezone(tzname)
            return [isodt.astimezone(tz)]
        except:  # pragma: no cover
            # Unlikely to occur, might do though if Project Haystack changes
            # its timezone list or if a system doesn't recognise a particular
            # timezone.
            return [isodt]  # Failed, leave alone
    else:
        return [isodt]


hs_dateTime = And([
    hs_isoDateTime,
    Optional(And([
        Suppress(Literal(' ')),
        hs_timeZoneName
    ]))
]).setParseAction(_parse_datetime)

# Quantities and raw numeric values
hs_unitChar = Or([
    hs_alpha,
    Word(u'%_/$' + u''.join([
        six.unichr(c)
        for c in range(0x0080, 0xffff)
    ]), exact=1)
])
hs_unit = Combine(OneOrMore(hs_unitChar))
hs_exp = Combine(And([
    CaselessLiteral('e'),
    Optional(hs_plusMinus),
    hs_digits
]))
hs_decimal = Combine(And([
    Optional(Literal('-')),
    hs_digits,
    Optional(And([
        Literal('.'),
        hs_digits
    ])),
    Optional(hs_exp)
])).setParseAction(lambda toks: [float(toks[0])])

hs_quantity = And([hs_decimal, hs_unit]).setParseAction(
    lambda toks: [Quantity(toks[0], unit=toks[1])])
hs_number = Or([
    hs_quantity,
    hs_decimal,
    Or([
        Literal('INF'),
        Literal('-INF'),
        Literal('NaN')
    ]).setParseAction(lambda toks: [float(toks[0])])
])

# URIs
hs_uriChar = Regex(r"([^\x00-\x1f\\`]|\\[bfnrt\\:/?" \
                   + r"#\[\]@&=;`]|\\[uU][0-9a-fA-F]{4})")
hs_uri = Combine(And([
    Suppress(Literal('`')),
    ZeroOrMore(hs_uriChar),
    Suppress(Literal('`'))
])).setParseAction(lambda toks: [Uri(_unescape(toks[0], uri=True))])

# Strings
hs_strChar = Regex(r"([^\x00-\x1f\\\"]|\\[bfnrt\\\"$]|\\[uU][0-9a-fA-F]{4})")
hs_str = Combine(And([
    Suppress(Literal('"')),
    ZeroOrMore(hs_strChar),
    Suppress(Literal('"'))
])).setParseAction(lambda toks: [_unescape(toks[0], uri=False)])

# References
hs_refChar = Or([hs_alpha, hs_digit, Word('_:-.~', exact=1)])
hs_ref = And([
    Suppress(Literal('@')),
    Combine(ZeroOrMore(hs_refChar)),
    Optional(And([
        Suppress(Literal(' ')),
        hs_str
    ]))
]).setParseAction(lambda toks: [ \
    Ref(toks[0], toks[1] if len(toks) > 1 else None) \
    ])

# Bins
hs_binChar = Regex(r"[\x20-\x27\x2a-\x7f]")
hs_bin = Combine(And([
    Suppress(Literal('Bin(')),
    Combine(ZeroOrMore(hs_binChar)),
    Suppress(Literal(')'))
])).setParseAction(lambda toks: [Bin(toks[0])])

# Haystack 3.0 XStr(...)
hs_xstr = And([
    Regex(r"[a-zA-Z0-9_]+"),
    Suppress(Literal('(')),
    hs_str,
    Suppress(Literal(')'))
]).setParseAction(lambda toks: [XStr(toks[0], toks[1])])

# Booleans
hs_bool = Word('TF', min=1, max=1, exact=1).setParseAction( \
    lambda toks: [toks[0] == 'T'])

# Singleton values
hs_remove = Literal('R').setParseAction( \
    lambda toks: [REMOVE]).setName('remove')
hs_marker = Literal('M').setParseAction( \
    lambda toks: [MARKER]).setName('marker')
hs_null = Literal('N').setParseAction( \
    lambda toks: [None]).setName('null')
hs_na = Literal('NA').setParseAction( \
    lambda toks: [NA]).setName('na')
# Lists, these will probably be in Haystack 4.0, so let's not
# assume a version.  There are three cases:
# - Empty list: [ {optional whitespace} ]
# - List *with* trailing comma: [ 1, 2, 3, ]
# - List without trailing comma: [ 1, 2, 3 ]
#
# We need to handle this trailing separator case.  That for now means
# that a NULL within a list *MUST* be explicitly given using the 'N'
# literal: we cannot support implicit NULLs as they are ambiguous.
hs_list = GenerateMatch( \
    lambda ver: Group(Or([ \
        Suppress(Regex(r'[ *]')), \
        And([ \
            Suppress(Regex(r'\[ *')), \
            Optional(DelimitedList( \
                hs_scalar[ver], \
                delim=hs_valueSep)), \
            Suppress(Optional(hs_valueSep)), \
            Suppress(Regex(r' *\]')) \
            ]) \
        ])).setParseAction(lambda toks: toks.asList()))
# Tag IDs
hs_id = Regex(r'[a-z][a-zA-Z0-9_]*').setName('id')

# Grid building blocks
hs_cell = GenerateMatch( \
    lambda ver: Or([Empty().copy().setParseAction(lambda toks: [None]), \
                    hs_scalar[ver]]).setName('cell'))

# Dict
# There are three cases:
# - Empty dict: { {optional whitespace} }
# - map with marker: { m }
# - dics: { k:1  ]
#
hs_tagmarker = hs_id

hs_tagpair = GenerateMatch(
    lambda ver: And([hs_id,
                     Suppress(Regex(r': *')),
                     hs_scalar[ver]
                     ])
        .setParseAction(lambda toks: tuple(toks[:2]))
        .setName('tagPair'))

hs_tag = GenerateMatch(
    lambda ver: Or([hs_tagmarker, hs_tagpair[ver]])
        .setName('tag'))

hs_tags = GenerateMatch(
    lambda ver: ZeroOrMore(Or([hs_tag[ver], \
                               Suppress(Regex(r'[ *]'))])) \
        .setName('tags'))


def to_dict(tokenlist):
    result = {}
    i = 0
    it = enumerate(tokenlist)
    for i, tok in it:
        if i < len(tokenlist) - 2 and tokenlist[i + 1] == ':':
            result[tokenlist[i]] = tokenlist[i + 2]
            next(it)
            next(it)
        else:
            if isinstance(tok, six.string_types):
                result[tok] = MARKER
            elif isinstance(tok, tuple):
                result[tok[0]] = tok[1]
            else:
                result[tok] = MARKER

    return result

# def to_dict(tokenlist):
#     result = {}
#     for i, tok in enumerate(tokenlist):
#         if isinstance(tok, six.string_types):
#             result[tok] = MARKER
#         else:
#             result[tok[0]] = tok[1]
#     return result


hs_dict = GenerateMatch(
    lambda ver: Or([
        Suppress(Regex(r'[ *]')),
        And([
            Suppress(Regex(r'{ *')),
            hs_tags[ver],
            Suppress(Regex(r' *}'))
        ])
    ])
        .setName("dict")
        .setParseAction(to_dict)
)

hs_inner_grid = GenerateMatch( \
    lambda ver: And([
        Suppress(Regex(r'<< *')),
        hs_grid[ver],
        Suppress(Regex(r' *>>')),
    ]))

# All possible scalar values, by Haystack version
hs_scalar_2_0 <<= Or([hs_ref, hs_bin, hs_str, hs_uri, hs_dateTime,
                      hs_date, hs_time, hs_coord, hs_number, hs_null, hs_marker,
                      hs_remove, hs_bool]).setName('scalar')
hs_scalar_3_0 <<= Or([hs_ref, hs_xstr, hs_str, hs_uri, hs_dateTime,
                      hs_date, hs_time, hs_coord, hs_number, hs_na, hs_null, hs_marker,
                      hs_remove, hs_bool, hs_list[VER_3_0], hs_dict[VER_3_0], hs_inner_grid[VER_3_0]]).setName('scalar')

hs_nl = Combine(And([Optional(Literal('\r')), Literal('\n')]))

hs_row = GenerateMatch( \
    lambda ver: Group(And([DelimitedList(hs_cell[ver], delim=hs_valueSep),
                           Suppress(Regex(r' *')),
                           Suppress(hs_nl)
                           ])).setName('row'))

hs_rows = GenerateMatch( \
    lambda ver: Group(ZeroOrMore(hs_row[ver])).setName("rows"))

hs_metaPair = GenerateMatch( \
    lambda ver: And([ \
        hs_id, \
        Suppress(And([ \
            ZeroOrMore(Literal(' ')), \
            Literal(':'), \
            ZeroOrMore(Literal(' ')) \
            ])), \
        hs_scalar[ver] \
        ]).setParseAction(lambda toks: [tuple(toks[:2])]).setName('metaPair'))
hs_metaMarker = hs_id.copy().setParseAction( \
    lambda toks: [(toks[0], MARKER)]).setName('metaMarker')
hs_metaItem = GenerateMatch( \
    lambda ver: Or([ \
        hs_metaMarker, \
        hs_metaPair[ver] \
        ]).setName('metaItem'))
hs_meta = GenerateMatch( \
    lambda ver: DelimitedList(hs_metaItem[ver], \
                              delim=' ').setParseAction( \
        lambda toks: [SortableDict(toks.asList())] \
        ).setName('meta'))

hs_col = GenerateMatch( \
    lambda ver: And([ \
        hs_id, \
        Optional(And([ \
            Suppress(Literal(' ')), \
            hs_meta[ver]
        ])).setName('colMeta') \
        ]).setParseAction(lambda toks: [ \
        (toks[0], toks[1] if len(toks) > 1 else {})]))

hs_cols = GenerateMatch( \
    lambda ver: And([
        DelimitedList(
            hs_col[ver], delim=hs_valueSep).setParseAction(  # + hs_nl
            lambda toks: [SortableDict(toks.asList())]),
        Suppress(Regex(r' *')),
        Suppress(hs_nl)
    ])
)

hs_gridVer = Combine(And([Suppress(Literal('ver:')) + hs_str]))


def _assign_ver(toks):
    ver = toks[0]
    if len(toks) > 1:
        grid_meta = toks[1]
    else:
        grid_meta = SortableDict()

    # Put 'ver' at the start
    grid_meta.add_item('ver', ver, index=0)
    return grid_meta


hs_gridMeta = GenerateMatch( \
    lambda ver: And([ \
        hs_gridVer, \
        Optional(And([ \
            Suppress(Literal(' ')), \
            hs_meta[ver] \
            ])).setName('gridMeta'),
        Suppress(Regex(r' *')),
        Suppress(hs_nl)
    ]).setParseAction(_assign_ver))  # + hs_nl


def _gen_grid(toks):
    (grid_meta, col_meta, rows) = toks
    if len(rows) == 1 and rows[0] == None:
        rows = []
    g = Grid(version=grid_meta.pop('ver'),
             metadata=grid_meta,
             columns=list(col_meta.items()))
    g.extend(map(lambda row: dict(zip(col_meta.keys(), row)), rows))
    return g


hs_grid_2_0 <<= And([ \
    hs_gridMeta[VER_2_0],
    hs_cols[VER_2_0],
    hs_rows[VER_2_0],
]).setParseAction(_gen_grid)

hs_grid_3_0 <<= And([ \
    hs_gridMeta[VER_3_0],
    hs_cols[VER_3_0],
    hs_rows[VER_3_0],
]).setParseAction(_gen_grid)


def parse_grid(grid_data, parseAll=True):
    """
    Parse the incoming grid.
    """
    try:
        # First element is the grid metadata
        ver_match = VERSION_RE.match(grid_data)
        if ver_match is None:
            raise ZincParseException(
                'Could not determine version from %r' % NEWLINE_RE.split(grid_data)[0],
                grid_data, 1, 1)
        version = Version(ver_match.group(1))

        # Now parse the grid of the grid accordingly
        g = hs_grid[version].parseString(grid_data, parseAll=parseAll)[0]
        return g
    except pp.ParseException as pe:
        LOG.debug('Failing grid: %r', grid_data)
        raise ZincParseException(
            'Failed to parse: %s' % reformat_exception(pe, pe.lineno),
            grid_data, pe.lineno, pe.col)
    except:
        LOG.debug('Failing grid: %r', grid_data)
        raise ZincParseException(
            'Failed to parse: %s' % sys.exc_info()[0], grid_data, 0, 0)


def parse_scalar(scalar_data, version):
    """
    Parse a Project Haystack scalar in ZINC format.
    """
    try:
        return hs_scalar[version].parseString(scalar_data, parseAll=True)[0]
    except pp.ParseException as pe:
        # Raise a new exception with the appropriate line number.
        raise ZincParseException(
            'Failed to parse scalar: %s' % reformat_exception(pe),
            scalar_data, 1, pe.col)
    except:
        LOG.debug('Failing scalar data: %r (version %r)',
                  scalar_data, version)
        raise



# -*- coding: utf-8 -*-
# Grid handling tests
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

# Assume unicode literals as per Python 3
from __future__ import unicode_literals

import copy
import datetime

from hszinc import Grid, Version, VER_3_0, Quantity, Coordinate
from hszinc.sortabledict import SortableDict


def test_grid_given_metadata():
    # Test that when passing in metadata, it is set as given.
    meta = SortableDict()
    meta['first'] = 1
    meta['second'] = 2
    meta['third'] = 3
    meta['fourth'] = 4

    assert list(Grid(metadata=meta).metadata.items()) == [
        ('first', 1), ('second', 2), ('third', 3), ('fourth', 4)
    ]


def test_grid_given_column_list():
    col_list = [('col1', [('c1m1', None), ('c1m2', None)]),
                ('col2', [('c2m1', None), ('c2m2', None)])]
    g = Grid(columns=col_list)
    assert list(g.column.keys()) == ['col1', 'col2']
    for col, meta in col_list:
        assert list(g.column[col].items()) == meta


def test_grid_given_column_dict():
    cols = SortableDict([('col1', [('c1m1', None), ('c1m2', None)]),
                         ('col2', [('c2m1', None), ('c2m2', None)])])
    g = Grid(columns=cols)
    assert list(g.column.keys()) == ['col1', 'col2']
    for col, meta in cols.items():
        assert list(g.column[col].items()) == meta


def test_grid_given_column_meta_dict():
    cols = SortableDict([('col1', SortableDict([('c1m1', None), ('c1m2', None)])),
                         ('col2', SortableDict([('c2m1', None), ('c2m2', None)]))])
    g = Grid(columns=cols)
    assert list(g.column.keys()) == ['col1', 'col2']
    for col, meta in cols.items():
        assert list(g.column[col].items()) == list(meta.items())


def test_grid_getitem():
    g = Grid()
    g.column['test'] = {}

    row = {'test': 'This is a test'}
    g.append(row)
    assert g[0] is row


def test_grid_setitem():
    g = Grid()
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = {'test': 'This is another test'}
    g.append(row_1)
    g[0] = row_2
    assert g[0] is row_2


def test_grid_append_notdict():
    g = Grid()
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = 'This is not a dict'
    g.append(row_1)
    try:
        g.append(row_2)
        assert False
    except TypeError as e:
        assert str(e) == 'value must be a dict'
        assert len(g) == 1


def test_grid_append_v2_list_fail():
    g = Grid(version='2.0')
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = {'test': ['This should fail']}
    g.append(row_1)
    try:
        g.append(row_2)
        assert False, 'Appended invalid data type'
    except ValueError as e:
        assert str(e) == 'Data type requires version 3.0'


def test_grid_append_nover_list():
    g = Grid()
    assert g.version == Version('2.0')
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = {'test': ['This should fail']}
    g.append(row_1)
    assert g.version == Version('2.0')
    g.append(row_2)
    assert g.version == Version('3.0')


def test_grid_setitem_notdict():
    g = Grid()
    g.column['test'] = {}

    row = {'test': 'This is a test'}
    g.append(row)

    try:
        g[0] = 'This is not a dict'
        assert False, 'Accepted a string'
    except TypeError:
        pass
    assert len(g) == 1
    assert g[0]['test'] == 'This is a test'


def test_grid_del():
    g = Grid()
    rows = [
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ]
    g.column['test'] = {}
    g.extend(rows)
    assert len(g) == 3
    del g[1]
    assert len(g) == 2
    assert g[0] is rows[0]
    assert g[1] is rows[2]


def test_grid_insert():
    g = Grid()
    rows = [
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ]
    g.column['test'] = {}
    g.extend(rows)
    assert len(g) == 3
    new_row = {'test': 'new'}
    g.insert(1, new_row)
    assert len(g) == 4
    assert g[0] is rows[0]
    assert g[1] is new_row
    assert g[2] is rows[1]
    assert g[3] is rows[2]


def test_grid_str():
    g = Grid(version=VER_3_0)
    rows = [
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ]
    g.column['test'] = {}
    g.extend(rows)
    assert repr(g) == '<Grid>\n' \
                      '\tVersion: 3.0\n' \
                      '\tColumns:\n' \
                      '\t\ttest\n' \
                      '\tRow    0:\n' \
                      '\ttest=1\n' \
                      '\tRow    1:\n' \
                      '\ttest=2\n' \
                      '\tRow    2:\n' \
                      '\ttest=3\n' \
                      '</Grid>'


def test_grid_equal():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    assert ref == copy.deepcopy(ref)


def test_grid_equal_with_complex_datas():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': datetime.datetime(2010, 11, 28, 7, 23, 2, 600000)},
        {'test': Quantity(500, 'kg')},
        {'test': Coordinate(100, 100)},
        {'test': 1.0},
    ])
    similar = Grid()
    similar.column['test'] = {}
    similar.extend([
        {'test': datetime.datetime(2010, 11, 28, 7, 23, 2, 500000)},
        {'test': Quantity(500.000001, 'kg')},
        {'test': Coordinate(100.000001, 100.000001)},
        {'test': 1.000001},
    ])

    assert ref == similar


def test_grid_not_equal_metadata():
    ref = Grid(metadata={"x": {}})
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff.metadata.append('add')
    assert ref != diff

    diff = copy.deepcopy(ref)
    diff.metadata["x"] = 1
    assert ref != diff


def test_grid_not_equal_col_with_new_metadata():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff.column.add_item('test', 'add')
    assert ref != diff


def test_grid_not_equal__with_new_col():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff.column["added"] = {}
    assert ref != diff


def test_grid_not_equal_col_with_updated_metadata():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff.column.pop('test')
    diff.column.add_item('new', 'add')
    assert ref != diff


def test_grid_not_equal_col_with_updated_metadata():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff.column['test']['test'] = None
    assert ref != diff


def test_grid_not_equal_with_updated_value():
    ref = Grid()
    ref.column['test'] = {"a": 1}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff.column['test'] = {"a": ""}
    assert ref != diff


def test_grid_new_row():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff.append({'test': 4})
    assert ref != diff


def test_grid_not_equal_row():
    ref = Grid()
    ref.column['test'] = {}
    ref.extend([
        {'test': 1},
        {'test': 2},
        {'test': 3}
    ])
    diff = copy.deepcopy(ref)
    diff[0] = {'test': 4}
    assert ref != diff


def test_grid_index():
    grid = Grid()
    grid.column['id'] = {}
    grid.column['val'] = {}
    grid.insert(0, {'id': 'idx1'})
    assert 'idx1' in grid._index
    assert grid['idx1']
    assert grid.get('idx1')
    grid.insert(1, {'id': 'idx2'})
    assert 'idx1' in grid._index
    assert 'idx2' in grid._index
    assert grid.get('idx2')
    grid[0] = {'id': 'idx3'}
    assert not 'idx1' in grid._index
    assert 'idx3' in grid._index
    assert grid.get('idx3')
    del grid[1]
    assert not 'idx2' in grid._index
    grid.extend([
        {'id': 'idx5'},
        {'id': 'idx6'},
    ])
    assert 'idx5' in grid._index
    assert 'idx6' in grid._index
    grid[0]['id'] = 'idx4'
    grid.reindex()
    assert 'idx4' in grid._index
    assert grid.get('idx4')


def test_slice():
    grid = Grid(columns={'id': {}, 'site': {}})
    grid.append({'id': 'id1', })
    grid.append({'id': 'id2'})
    grid.append({'id': 'id3'})

    result = grid[0:2]
    assert len(result) == 2
    assert result['id1']
    assert result['id2']
